# vim:fileencoding=utf-8
# Copyright (c) 2009 gocept gmbh & co. kg
# See also LICENSE.txt

import gocept.directory.app
import gocept.directory.testing
import os.path
import transaction
import unittest
import xmlrpclib
import zope.app.testing.functional
import zope.app.testing.xmlrpc


class TestRing0(zope.app.testing.functional.FunctionalTestCase):

    layer = gocept.directory.testing.DataLayer

    def setUp(self):
        super(TestRing0, self).setUp()
        self.directory = directory = self.getRootFolder()['directory']
        self.old = zope.app.component.hooks.getSite()
        zope.app.component.hooks.setSite(self.directory)

        newyork = gocept.directory.location.Location()
        newyork.abbreviation = 'ny'
        directory['newyork'] = newyork
        for i, vlan in [(1, 'ipmi'), (1, 'mgm'),
                        (2, 'fe'), (3, 'srv'), (4, 'sto')]:
            n = gocept.directory.network.Network()
            n.cidr = '231.20.%s.0/24' % i
            newyork['vlan-%s' % vlan][n.cidr.replace('/', '-')] = n

            if vlan != 'ipmi':
                n = gocept.directory.network.Network()
                n.cidr = '2002:470:9aaf:4%s::/64' % i
                newyork['vlan-%s' % vlan][n.cidr.replace('/', '-')] = n

        jamaica = gocept.directory.location.Location()
        jamaica.abbreviation = 'jamaica'
        directory['jamaica'] = jamaica
        for i, vlan in [(1, 'ipmi'), (1, 'mgm'),
                        (2, 'fe'), (3, 'srv')]:
            n = gocept.directory.network.Network()
            n.cidr = '172.20.%s.0/24' % (i)
            jamaica['vlan-%s' % vlan][n.cidr.replace('/', '-')] = n
        for i, vlan in [(4, 'sto'), (3, 'srv'), (2, 'fe')]:
            n = gocept.directory.network.Network()
            n.cidr = '2001:470:9aaf:4%s::/64' % i
            jamaica['vlan-%s' % vlan][n.cidr.replace('/', '-')] = n

        services = directory['rg-services']
        self.stampy = stampy = gocept.directory.nodes.PhysicalMachineFactory()(
            'stampy', 4099, newyork, services)
        stampy.directory_ring = 0
        stampy.directory_password = 'gfhdjk'
        stampy['mgm'].mac = '00:1e:c9:ad:4a:a6'
        newyork['vlan-mgm']['231.20.1.0-24'].allocate(stampy['mgm'])

        stampy['ipmi'].mac = '00:1e:c9:ad:4a:a7'
        newyork['vlan-ipmi']['231.20.1.0-24'].allocate(stampy['ipmi'])

        stampy['fe'].mac = '00:15:17:91:d2:f0'
        newyork['vlan-fe']['231.20.2.0-24'].allocate(stampy['fe'])

        stampy['srv'].mac = '00:1e:c9:ad:4a:a0'
        newyork['vlan-srv']['231.20.3.0-24'].allocate(stampy['srv'])

        stampy['sto'].mac = '00:1e:c9:ad:4a:a2'
        newyork['vlan-sto']['231.20.4.0-24'].allocate(stampy['sto'])

        test00 = gocept.directory.nodes.VirtualMachine()
        test00.directory_password = None
        test00.directory_ring = 1
        test00.name = 'test00'
        test00.location = newyork
        test00.memory = 512
        directory['rg-services']['test00'] = test00

        self.rpc = gocept.directory.testing.ServerProxy(
            'http://localhost/directory/', user='stampy', password='gfhdjk')

    def tearDown(self):
        zope.app.component.hooks.setSite(self.old)
        super(TestRing0, self).tearDown()

    def test_security(self):
        self.rpc = gocept.directory.testing.ServerProxy(
            'http://localhost/directory/')
        self.assertRaises(
            xmlrpclib.ProtocolError, self.rpc.get_password, 'foobar')

    def test_password(self):
        self.assertRaises(xmlrpclib.Fault, self.rpc.get_password, 'ldap')
        ldap = gocept.directory.password.Password()
        ldap.clear_text = 'nenpotgaf'
        self.directory['pw-ldap'] = ldap
        self.assertEquals('nenpotgaf', self.rpc.get_password('ldap'))

    def test_lookup_node_no_directory_password(self):

        self.assertRaises(
            KeyError,
            lambda: self.rpc.lookup_node('test00')
                    ['parameters']['directory_password'])

        self.directory.node_by_name('test00').directory_password = 'asdf'
        self.assertEquals('asdf',
                          self.rpc.lookup_node('test00')
                              ['parameters']['directory_password'])

    def test_lookup_node_no_kvm_host(self):
        self.directory.node_by_name('test00').kvm_host = None
        self.assertEquals(
            'none', self.rpc.lookup_node('test00')['parameters']['kvm_host'])

    def test_lookup_node(self):
        stampy = self.directory.node_by_name('stampy')
        stampy.puppet_classes = ['role::nagios',
                                 'role::portage_binhost',
                                 'role::puppetmaster']
        stampy.environment = 'testing'

        self.assertEquals({
            'classes': ['role::nagios',
                        'role::portage_binhost',
                        'role::puppetmaster'],
            'name': 'stampy',
            'parameters': {
                'directory_password': 'gfhdjk',
                'directory_ring': 0,
                'environment': 'testing',
                'machine': 'physical',
                'location': 'ny',
                'mac_ipmi': '00:1e:c9:ad:4a:a7',
                'mac_mgm': '00:1e:c9:ad:4a:a6',
                'mac_fe': '00:15:17:91:d2:f0',
                'mac_srv': '00:1e:c9:ad:4a:a0',
                'mac_sto': '00:1e:c9:ad:4a:a2',
                'resource_group': 'services',
                'ip4_mgm': 2,
                'ip4_ipmi': 3,
                'ip4_fe': 2,
                'ip4_srv': 2,
                'ip4_sto': 2,
                'id': 4099}},
            self.rpc.lookup_node('stampy'))

    def test_lookup_invalid_node_should_fail(self):
        self.assertRaises(xmlrpclib.Fault, self.rpc.lookup_node, 'foobar')

    def test_list_virtual_machines(self):
        vms = self.rpc.list_virtual_machines()
        self.assertEquals(1, len(vms))
        self.assertEquals('test00', vms[0]['name'])

    def test_list_networks(self):
        self.assertEquals(
            ['172.20.1.0/24', '172.20.2.0/24', '172.20.3.0/24',
             '2001:470:9aaf:42::/64',
             '2001:470:9aaf:43::/64', '2001:470:9aaf:44::/64',
             '2002:470:9aaf:41::/64', '2002:470:9aaf:42::/64',
             '2002:470:9aaf:43::/64', '2002:470:9aaf:44::/64',
             '231.20.1.0/24', '231.20.2.0/24', '231.20.3.0/24',
             '231.20.4.0/24',
             ],
            self.rpc.list_networks())

    def test_list_nodes(self):
        nodes = self.rpc.list_nodes()
        self.assertEquals(['stampy', 'test00'],
                          sorted([n['name'] for n in nodes]))

    def test_list_locations(self):
        self.assertEquals(
            ['jamaica', 'ny'],
            self.rpc.list_locations())

    def test_lookup_networks(self):
        self.assertEquals(
            {'sto': ['2001:470:9aaf:44::/64'],
             'srv': ['172.20.3.0/24', '2001:470:9aaf:43::/64'],
             'fe': ['172.20.2.0/24', '2001:470:9aaf:42::/64'],
             'mgm': ['172.20.1.0/24'],
             'ipmi': ['172.20.1.0/24']},
            self.rpc.lookup_networks('jamaica'))

    def test_lookup_networks_with_af(self):
        self.assertEquals(
            {'srv': ['172.20.3.0/24'],
             'fe': ['172.20.2.0/24'],
             'mgm': ['172.20.1.0/24'],
             'ipmi': ['172.20.1.0/24']},
            self.rpc.lookup_networks('jamaica', 'ipv4'))

    def test_lookup_interfaces(self):
        for addr in self.stampy['fe'].addresses:
            addr.reverse = u'ns1.gocept.com.'
        self.assertEquals(
            {'mgm': [['2002:470:9aaf:41::2/64', None],
                     ['231.20.1.2/24', None]],
             'ipmi': [['231.20.1.3/24', None]],
             'fe': [['2002:470:9aaf:42::2/64', u'ns1.gocept.com.'],
                    ['231.20.2.2/24', u'ns1.gocept.com.']],
             'srv': [['2002:470:9aaf:43::2/64', None],
                     ['231.20.3.2/24', None]],
             'sto': [['2002:470:9aaf:44::2/64', None],
                     ['231.20.4.2/24', None]]},
            self.rpc.lookup_interfaces('stampy'))

    def test_lookup_interfaces_with_af(self):
        self.assertEquals(
            {'fe': [['231.20.2.2/24', None]],
             'srv': [['231.20.3.2/24', None]],
             'mgm': [['231.20.1.2/24', None]],
             'ipmi': [['231.20.1.3/24', None]],
             'sto': [['231.20.4.2/24', None]]},
            self.rpc.lookup_interfaces('stampy', 'ipv4'))

    def test_lookup_addresses(self):
        self.assertEquals(
            sorted(['2002:470:9aaf:41::2/64', '231.20.1.2/24',
             '2002:470:9aaf:44::2/64', '231.20.4.2/24',
             '2002:470:9aaf:43::2/64', '231.20.3.2/24',
             '2002:470:9aaf:42::2/64', '231.20.2.2/24',
             '231.20.1.3/24']),
            sorted(self.rpc.lookup_addresses('stampy')))

    def test_lookup_addresses_with_vlan(self):
        self.assertEquals(
            ['2002:470:9aaf:42::2/64', '231.20.2.2/24'],
            self.rpc.lookup_addresses('stampy', 'fe'))

    def test_lookup_addresses_with_alias_vlan(self):
        self.assertEquals(
            ['231.20.1.3/24'],
            self.rpc.lookup_addresses('stampy', 'ipmi'))
        self.assertEquals(
            ['2002:470:9aaf:41::2/64', '231.20.1.2/24'],
            self.rpc.lookup_addresses('stampy', 'mgm'))

    def test_lookup_addresses_with_af(self):
        self.assertEquals(
            ['2002:470:9aaf:41::2/64', '2002:470:9aaf:42::2/64',
             '2002:470:9aaf:43::2/64', '2002:470:9aaf:44::2/64'],
            sorted(self.rpc.lookup_addresses('stampy', '', 'ipv6')))

    def test_consistency(self):
        self.assertEquals(
            [], self.rpc.check_consistency())

    def test_consistency_duplicate_global_id(self):
        self.directory.node_by_name('stampy').global_id = 1
        self.directory.node_by_name('test00').global_id = 1
        self.assertEquals(
            ['Duplicate node ID: 1 by stampy, test00'],
            self.rpc.check_consistency())

    def test_consistency_duplicate_address(self):
        nic1 = self.directory.node_by_name('stampy')['ipmi']
        nic2 = self.directory.node_by_name('stampy')['mgm']

        addr1 = list(nic1.addresses)[0]

        addr = gocept.directory.network.Address(addr1.ip)
        addr.resource = nic2

        nic2.vlan[addr1.network.__name__][addr.ip] = addr
        nic2.addresses.add(addr)

        self.assertEquals(
            ['Duplicate IP: 231.20.1.3 by stampy.ipmi, stampy.mgm'],
            self.rpc.check_consistency())

    def test_consistency_duplicate_mac(self):
        nic1 = self.directory.node_by_name('stampy')['ipmi']
        nic2 = self.directory.node_by_name('stampy')['mgm']

        nic2.mac = nic1.mac

        self.assertEquals(
            ['Duplicate MAC: 00:1e:c9:ad:4a:a7 by stampy.ipmi, stampy.mgm'],
            self.rpc.check_consistency())

    def test_regular_lookup_node_access_denied(self):
        # Turning stampy into a ring 1 node causes him not being able to
        # acecss the lookup_node function anymore:
        self.directory.node_by_name('stampy').directory_ring = 1
        self.assertRaises(xmlrpclib.ProtocolError,
                          self.rpc.lookup_node, 'stampy')

    def test_list_resource_groups(self):
        # Set up another RG that will also be shown
        self.directory['rg-foo'] = (
            gocept.directory.resourcegroup.ResourceGroup())
        self.directory['rg-foo'].name = 'foo'
        self.assertEquals(
            ['foo', 'services'], self.rpc.list_resource_groups())
