# Copyright (c) 2008-2010 gocept gmbh & co. kg
# See also LICENSE.txt

import re
import grok
import zc.sourcefactory.basic
import zc.sourcefactory.contextual
import zope.interface
import zope.schema
import zope.component
import zope.app.component.hooks
import zope.app.form.interfaces


class IResource(zope.interface.Interface):

    title = zope.interface.Attribute(
        'A human-readable title for the resource.')


class IDirectory(zope.interface.Interface):
    """The directory."""

    locations = zope.interface.Attribute('All locations')
    vlan_classes = zope.interface.Attribute('All VLAN classes')


class IResourceGroup(zope.interface.Interface):
    """A group of resources."""

    name = zope.schema.ASCIILine(title=u'Name')
    nodes = zope.interface.Attribute(
        "All nodes that belong to this resource group.")


class ILocation(zope.interface.Interface):

    title = zope.schema.TextLine(title=u'Title')
    abbreviation = zope.schema.ASCIILine(title=u'Short name / Identifier')

    vlans = zope.interface.Attribute('The VLANs of this location.')


class IVLANClass(zope.interface.Interface):

    title = zope.schema.TextLine(title=u'Titel')
    short = zope.schema.ASCIILine(title=u'Kurzname')
    description = zope.schema.TextLine(title=u'Description')
    number = zope.schema.Int(title=u'Number')
    color = zope.schema.ASCIILine(title=u'Color (hex)')


class IVLAN(zope.interface.Interface):
    """A network segment in a location."""

    class_ = zope.interface.Attribute('The VLAN class this vlan belongs to.')

    networks = zope.interface.Attribute(
        'The networks reachable on this VLAN.')


class INetwork(zope.interface.Interface):
    """An IP network."""

    cidr = zope.schema.ASCIILine(
        title=u'Network',
        description=u'Please use a CIDR representation of the network.')

    def allocate(resource, address=None):
        """Allocate an address for the given resource.

        If the address has already been allocated a KeyError will be raised.

        If the address is `None` then an address will be automatically picked.

        Assumes that the resource has a set attribute `addresses` where the
        address will be registered in.

        Warning: that the address object might be substituted with an already
        existing address object that has the same value.

        """


class IAddress(zope.interface.Interface):
    """An IP address."""

    network = zope.interface.Attribute('The network this address belongs to.')
    resource = zope.interface.Attribute(
        'The resource this address is being used by.')
    ip = zope.interface.Attribute('The IP this address represents.')
    index = zope.interface.Attribute(
        'The index the address has in its network')
    reverse = zope.schema.TextLine(title=u'Reverse DNS name (optional)',
        description=u'FQDNs need to have a trailing dot; leave out trailing '
                    u'dot for relative domain names',
        required=False)


class NICNetworks(zc.sourcefactory.contextual.BasicContextualSourceFactory):

    def getValues(self, nic):
        for network in nic.vlan.networks:
            yield network

    def getTitle(self, nic, network):
        return network.cidr


class LocationNetworks(
        zc.sourcefactory.contextual.BasicContextualSourceFactory):

    def getValues(self, location):
        for vlan in location.vlans:
            for network in vlan.networks:
                yield network

    def getTitle(self, location, network):
        return u'%s %s.%s' % (network.cidr, network.vlan.class_.short,
                              network.vlan.location.abbreviation)


class NICVLANs(zc.sourcefactory.contextual.BasicContextualSourceFactory):

    def getValues(self, nic):
        if INode.providedBy(nic):
            node = nic
        else:
            node = nic.node
        return node.location.vlans

    def getTitle(self, nic, vlan):
        return vlan.class_.title


class INIC(zope.interface.Interface):
    """A network interface card."""

    mac = zope.schema.ASCIILine(title=u'MAC address')
    vlan = zope.schema.Choice(source=NICVLANs())
    addresses = zope.interface.Attribute(
        u'Addresses associated with this NIC.')


class Locations(zc.sourcefactory.contextual.BasicContextualSourceFactory):

    def getValues(self, context):
        directory = zope.app.component.hooks.getSite()
        return directory.locations

    def getTitle(self, context, location):
        return location.title


class PuppetClasses(zc.sourcefactory.basic.BasicSourceFactory):

    def getValues(self):
        roles = (zope.app.appsetup.product.getProductConfiguration(
                 'gocept.directory')['puppet-roles'])
        return [x.strip() for x in open(roles)]


class PuppetEnvironments(zc.sourcefactory.basic.BasicSourceFactory):

    def getValues(self):
        env = (zope.app.appsetup.product.getProductConfiguration(
                 'gocept.directory')['puppet-environments'])
        return [x.strip() for x in open(env)]


re_hostname_characters = re.compile(r'^[a-zA-Z0-9][a-zA-Z0-9-]{0,62}$')


def check_hostname(value):
    """Return True if the hostname is valid"""
    return re_hostname_characters.match(value) and not value.isdigit()


class DuplicateGlobalIdError(zope.schema.ValidationError):
    u"""This global ID is already in use."""
    zope.interface.implements(zope.app.form.interfaces.IWidgetInputError)


def check_global_id_unique(context):
    # Avoid circular import.
    from gocept.directory.app import locate_directory_from
    directory = locate_directory_from(zope.app.component.hooks.getSite())
    # We can't compare object identity here as context is always just a
    # FormData instance. However, we rely on name to provide uniqueness by
    # itself.
    existing = [node for node in directory.nodes
                if node.global_id == context.global_id]
    existing = [x for x in existing
                if x.name != context.name]
    if existing:
        raise DuplicateGlobalIdError()


class INode(zope.interface.Interface):

    class_ = zope.interface.Attribute('Type of node: virtual or physical')
    nics = zope.interface.Attribute('List of NICs for this node.')
    resource_group = zope.interface.Attribute(
        u'The resource group this node belongs to')

    name = zope.schema.ASCIILine(title=u'Hostname',
                                 constraint=check_hostname,
                                 readonly=True)
    global_id = zope.schema.Int(title=u'Global node ID',
                                min=4100, max=65535, readonly=True)
    global_id_unique = zope.interface.invariant(
        check_global_id_unique)
    location = zope.schema.Choice(
        title=u'Location',
        source=Locations())

    puppet_classes = zope.schema.Set(
        title=u'Puppet-Klassen',
        value_type=zope.schema.Choice(source=PuppetClasses()))

    environment = zope.schema.Choice(title=u'Environment',
                                     source=PuppetEnvironments())

    # XXX Password should be automatically generated
    directory_password = zope.schema.TextLine(
        title=u'The directory API password for this node')
    directory_ring = zope.schema.Int(
        title=u'Access level (ring) for directory API')


class KVMHosts(zc.sourcefactory.contextual.BasicContextualSourceFactory):

    def getValues(self, context):
        yield None
        directory = zope.app.component.hooks.getSite()
        for node in directory.nodes:
            if 'role::kvm_host' in node.puppet_classes:
                yield node

    def getTitle(self, context, node):
        if node is None:
            return u'--disabled--'
        return node.name

    def getToken(self, context, node):
        if node is None:
            return '.none'
        return super(KVMHosts, self).getToken(context, node)


class IVirtualMachine(INode):

    location = zope.interface.Attribute(
        'Override schema-field of INode. Location is derived from kvm_host')

    # XXX This is computed and storage-implementation specific. Do this on
    # client?
    rootdev = zope.interface.Attribute('Name of the root device')

    memory = zope.schema.Int(title=u'Memory in MiB')

    # XXX setting the kvm host determines the location automatically. Once a
    # location has been set the host can not be changed to a different
    # location as the data would be on an unreachable storage server
    # otherwise.
    kvm_host = zope.schema.Choice(title=u'KVM host', source=KVMHosts(),
                                  required=False)


class IPassword(zope.interface.Interface):

    clear_text = zope.schema.TextLine(
        title=u'The clear text representation of the password')


class IConsistencyCheck(zope.interface.Interface):
    """Provides a consistency check for a context.

    Produces a list of errors when inconsistencies are found.

    Register objects as subscribers to this interface to be picked up by the
    global consistency check mechanisms.

    """

    errors = zope.interface.Attribute('The list of errors.')
