# Copyright (c) 2010 gocept gmbh & co. kg
# See also LICENSE.txt

import gocept.cache.method
import gocept.directory.app
import gocept.directory.consistency
import gocept.directory.form
import gocept.directory.interfaces
import gocept.directory.nic
import grok
import grok.interfaces
import megrok.pagelet
import netaddr
import os.path
import random
import string
import yaml
import zope.app.component.hooks
import zope.app.form.browser.source
import zope.component


class Node(grok.Container):

    grok.implements(gocept.directory.interfaces.INode,
                    gocept.directory.interfaces.IResource)

    puppet_classes = ()
    environment = 'production'

    location = gocept.reference.Reference(
        ensure_integrity=True, __name__='location')

    def __init__(self, name=None, global_id=None, location=None):
        super(Node, self).__init__()
        self.name = name
        self.global_id = global_id
        self.location = location

    @property
    def resource_group(self):
        return self.__parent__

    @property
    def title(self):
        return self.name

    @property
    def nics(self):
        for obj in self.values():
            if gocept.directory.interfaces.INIC.providedBy(obj):
                yield obj


@grok.subscribe(gocept.directory.interfaces.INode,
        grok.IObjectAddedEvent)
def node_added(node, event):
    directory = gocept.directory.app.locate_directory_from(node)
    directory.getSiteManager().registerUtility(node,
            provided=gocept.directory.interfaces.INode,
            name=node.__name__)


@grok.subscribe(gocept.directory.interfaces.INode,
        grok.IObjectRemovedEvent)
def node_removed(node, event):
    directory = gocept.directory.app.locate_directory_from(event.oldParent)
    assert directory.getSiteManager().unregisterUtility(
        node, provided=gocept.directory.interfaces.INode, name=event.oldName)


class PhysicalMachine(Node):

    class_ = 'physical'


class VirtualMachine(Node):

    grok.implements(gocept.directory.interfaces.IVirtualMachine)

    class_ = 'virtual'

    kvm_host = gocept.reference.Reference(ensure_integrity=True)

    def __init__(self, *args, **kw):
        super(VirtualMachine, self).__init__(*args, **kw)
        self.kvm_host = None

    @property
    def rootdev(self):
        return '/dev/iscsi/vm.%s-root' % self.name


class EditPhysical(gocept.directory.form.EditForm):

    grok.name('edit')
    grok.context(PhysicalMachine)
    form_fields = grok.AutoFields(gocept.directory.interfaces.INode)


class EditVirtual(gocept.directory.form.EditForm):

    grok.name('edit')
    grok.context(VirtualMachine)
    form_fields = grok.AutoFields(gocept.directory.interfaces.IVirtualMachine)


class ViewPhysical(megrok.pagelet.Pagelet):

    grok.require('gocept.directory.ManageDirectory')
    grok.name('index')
    grok.template('index_physical')
    grok.context(PhysicalMachine)


class ViewVirtual(megrok.pagelet.Pagelet):

    grok.require('gocept.directory.ManageDirectory')
    grok.name('index')
    grok.template('index_virtual')
    grok.context(VirtualMachine)


class DeleteConfirmation(megrok.pagelet.Pagelet):
    grok.require('gocept.directory.ManageDirectory')
    grok.name('delete')
    grok.template('delete')
    grok.context(Node)


class AddPhysicalMachine(gocept.directory.form.AddForm):

    grok.context(gocept.directory.interfaces.IResourceGroup)
    form_fields = (grok.AutoFields(gocept.directory.interfaces.INode).
                   omit('addresses', 'directory_password'))

    def __init__(self, context, request):
        super(AddPhysicalMachine, self).__init__(context, request)
        if 'form.global_id' not in self.request.form:
            # XXX duplicated in AddVirtualMachine
            directory = gocept.directory.app.locate_directory_from(context)
            self.request.form['form.global_id'] = max(
                gocept.directory.interfaces.INode['global_id'].min,
                max(x.global_id for x in directory.nodes) + 1)
        if ('form.directory_ring' not in self.request.form):
            self.request.form['form.directory_ring'] = 0

    def create(self, **data):
        factory = zope.component.getUtility(
            zope.component.interfaces.IFactory,
            'gocept.directory.PhysicalMachine')
        return factory(data['name'], data['global_id'], data['location'],
                       self.context)

    def add(self, obj):
        pass


class AddVirtualMachine(gocept.directory.form.AddForm):

    grok.context(gocept.directory.interfaces.IResourceGroup)
    form_fields = (grok.AutoFields(gocept.directory.interfaces.INode).
                   omit('addresses', 'directory_password'))

    def __init__(self, context, request):
        super(AddVirtualMachine, self).__init__(context, request)
        if 'form.global_id' not in self.request.form:
            # XXX Duplicated in AddPhysicalMachine
            directory = gocept.directory.app.locate_directory_from(context)
            if len(list(directory.nodes)):
                new_global_id = max(x.global_id for x in directory.nodes) + 1
            else:
                new_global_id = gocept.directory.interfaces.INode['global_id'].min
            self.request.form['form.global_id'] = new_global_id
        if ('form.name' not in self.request.form and
                hasattr(context, 'suggest_hostname')):
            self.request.form['form.name'] = context.suggest_hostname()
        if ('form.directory_ring' not in self.request.form):
            self.request.form['form.directory_ring'] = 1

    def create(self, **data):
        factory = zope.component.getUtility(
            zope.component.interfaces.IFactory,
            'gocept.directory.VirtualMachine')
        return factory(data['name'], data['global_id'], data['location'],
                       self.context)

    def add(self, obj):
        pass


class AddNIC(gocept.directory.form.BasicAddForm):

    factory = gocept.directory.nic.NIC

    grok.context(PhysicalMachine)
    form_fields = (grok.AutoFields(gocept.directory.interfaces.INIC).
                   omit('addresses'))


class MachineFactory(object):

    def __call__(self, name, id, location, resource_group):
        nodelabel = name.lower()
        othernode = zope.component.queryUtility(
                gocept.directory.interfaces.INode, nodelabel)
        if othernode:
            raise zope.exceptions.interfaces.DuplicationError(
                    'node "%s" already exists' % name)
        node = self.klass(name, id, location)
        resource_group[nodelabel] = node
        return node

    def create_nics(self, node, *vlans):
        for vlan in vlans:
            vlan = node.location['vlan-%s' % vlan]
            nic = gocept.directory.nic.NIC()
            nic.vlan = vlan
            node[vlan.class_.short] = nic
            node[vlan.class_.short].mac = str(netaddr.EUI(
                netaddr.EUI('02:00:00:%02x:00:00' % vlan.class_.number).value
                + int(node.global_id))).replace('-', ':')
            for net in vlan.networks:
                # Allocate IPv6 addresses
                if net.netaddr.version != 6:
                    continue
                net.allocate(nic)
            yield nic


class PhysicalMachineFactory(grok.GlobalUtility, MachineFactory):

    grok.implements(zope.component.interfaces.IFactory)
    grok.name('gocept.directory.PhysicalMachine')
    klass = PhysicalMachine

    def __call__(self, *args):
        node = super(PhysicalMachineFactory, self).__call__(*args)
        node.directory_password = (
            ''.join(random.choice(string.letters + string.digits) for x in
                    range(20)))
        for nic in self.create_nics(node, 'srv', 'fe', 'mgm', 'sto', 'ipmi'):
            # These are NICs with real MAC addresses, remove the computed ones
            nic.mac = ''
        return node


class VirtualMachineFactory(grok.GlobalUtility, MachineFactory):

    grok.implements(zope.component.interfaces.IFactory)
    grok.name('gocept.directory.VirtualMachine')
    klass = VirtualMachine

    def __call__(self, *args):
        node = super(VirtualMachineFactory, self).__call__(*args)
        node.memory = 256
        node.directory_password = (
            ''.join(random.choice(string.letters + string.digits) for x in
                    range(20)))

        list(self.create_nics(node, 'fe', 'srv'))
        return node


class GlobalIDConsistency(gocept.directory.consistency.UniqueResource):

    type_ = 'node ID'

    def resources(self):
        for node in self.directory.nodes:
            yield node.global_id, node.name
