# Copyright (c) 2010 gocept gmbh & co. kg
# See also LICENSE.txt

import unittest
import gocept.directory.app
import gocept.directory.location
import gocept.directory.nodes
import gocept.directory.testing
import zope.app.testing.functional
import zope.exceptions.interfaces


class NodeTests(zope.app.testing.functional.FunctionalTestCase):

    layer = gocept.directory.testing.DataLayer

    def setUp(self):
        super(NodeTests, self).setUp()
        self.directory = self.getRootFolder()['directory']
        self.oldsite = zope.app.component.hooks.getSite()
        zope.app.component.hooks.setSite(self.directory)
        self.directory['rg'] = self.rg = \
                gocept.directory.resourcegroup.ResourceGroup()
        self.directory['whq'] = self.location = \
                gocept.directory.location.Location()

    def tearDown(self):
        zope.app.component.hooks.setSite(self.oldsite)
        super(NodeTests, self).tearDown()

    def test_node_is_local_utility(self):
        self.directory['node1'] = \
                gocept.directory.nodes.Node('Node1')
        node = zope.component.getUtility(
                gocept.directory.interfaces.INode, name='node1')
        self.assertEquals(node.__name__, 'node1')

    def test_node_delete_removes_utility(self):
        self.directory['node1'] = \
                gocept.directory.nodes.Node('Node1')
        node = zope.component.getUtility(
                gocept.directory.interfaces.INode, name='node1')
        self.assertEquals(node.__name__, 'node1')
        del self.directory['node1']
        self.assertRaises(zope.component.ComponentLookupError,
                          zope.component.getUtility,
                          gocept.directory.interfaces.INode, name='node1')

    def test_global_id_consistency_no_nodes(self):
        check = gocept.directory.nodes.GlobalIDConsistency(self.directory)
        self.assertEquals([], check.errors)

    def test_global_id_consistency_one_node(self):
        self.rg['node1'] = node1 = gocept.directory.nodes.Node()
        node1.global_id = 1
        node1.name = 'node1'

        check = gocept.directory.nodes.GlobalIDConsistency(self.directory)
        self.assertEquals([], check.errors)

    def test_global_id_consistency_two_nodes_no_conflict(self):
        self.rg['node1'] = node1 = gocept.directory.nodes.Node()
        node1.global_id = 1
        node1.name = 'node1'

        self.rg['node2'] = node2 = gocept.directory.nodes.Node()
        node2.global_id = 2
        node2.name = 'node2'

        check = gocept.directory.nodes.GlobalIDConsistency(self.directory)
        self.assertEquals([], check.errors)

    def test_global_id_consistency_two_nodes_conflict(self):
        self.rg['node1'] = node1 = gocept.directory.nodes.Node()
        node1.name = 'node1'
        node1.global_id = 1

        self.rg['node2'] = node2 = gocept.directory.nodes.Node()
        node2.name = 'node2'
        node2.global_id = 1

        check = gocept.directory.nodes.GlobalIDConsistency(self.directory)
        self.assertEquals(['Duplicate node ID: 1 by node1, node2'],
                          check.errors)

    def test_vm_id_equals_name(self):
        factory = gocept.directory.nodes.VirtualMachineFactory()
        vm = factory('vmname', 4100, self.location, self.rg)
        self.assertEquals(vm.name, 'vmname')
        self.failUnless('vmname' in self.rg)

    def test_vm_lowercase_hostname_is_unique(self):
        factory = gocept.directory.nodes.VirtualMachineFactory()
        factory('Foo', 4100, self.location, self.rg)
        self.assertRaises(zope.exceptions.interfaces.DuplicationError,
                factory, 'fOO', 4101, self.location, self.rg)

    def test_vm_hostname_is_globally_unique(self):
        factory = gocept.directory.nodes.VirtualMachineFactory()
        factory('foo', 4100, self.location, self.rg)
        self.directory['rg2'] = rg2 = \
                gocept.directory.resourcegroup.ResourceGroup()
        self.assertRaises(zope.exceptions.interfaces.DuplicationError,
                factory, 'foo', 4101, self.location, rg2)

    def test_vm_gets_directory_password(self):
        factory = gocept.directory.nodes.VirtualMachineFactory()
        node = factory('foo', 4100, self.location, self.rg)
        self.assertNotEquals('', node.directory_password)

    def test_physical_machine_gets_directory_password(self):
        factory = gocept.directory.nodes.PhysicalMachineFactory()
        node = factory('foo', 4100, self.location, self.rg)
        self.assertNotEquals('', node.directory_password)
